# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

# This folder that contains tests which ensure that the results are almost identical
# to the results of the neuron toolbox.
